/**
 * ----------------------------------------------------------------------------------
 *
 * Copyright (C) 2009 runlevel3 GmbH
 *
 * ----------------------------------------------------------------------------------
 *
 * LICENSE:
 *
 * This file is part of CyberClockControl.
 *
 * CyberClockControl is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * CyberClockControl is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CyberClockControl; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * ----------------------------------------------------------------------------------
 *
 * Qt Creator / Qt 4.5.0
 *
 * ----------------------------------------------------------------------------------
 *
 * @copyright 2009 runlevel3 GmbH
 * @license   http://www.gnu.org/licenses/gpl.txt GNU General Public License v2
 * @link      http://www.runlevel3.de runlevel3
 **/


#include "math.h"
#include <QFile>
#include <QIntValidator>
#include <qfiledialog.h>
#include "mainwindow.h"
#include "ui_mainwindow.h"

#include "i2c_usb.h"

#define f_max       200000000           // 200 MHz
#define CAPLOAD_DEFAULT     64
#define XCODRV_DEFAULT      2


pll_settings cy_pll;
double Q, PB, PO, P, Fpll, Fref;
int div1n, div2n;
int f_pll_min, f_pll_max;
bool firstconnect = true;
bool do_not_change_sem = false;
extern bool chip_found;
char config_bob[256];
QString filename;

MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent), ui(new Ui::MainWindow)
{
    ui->setupUi(this);
    do_usb_ui_init(ui);
    Fref = ui->sbFreference->value();
    f_pll_min = 100000000;	// 100 MHz
    f_pll_max = 400000000;	// 400 MHz
    config_bob[0x13] = CAPLOAD_DEFAULT;
    config_bob[0x12] |= XCODRV_DEFAULT << 3;
    cy_pll.f1 = 0.0;
    cy_pll.f2 = 0.0;
    fillCheckboxes();
    fillComboboxes();
    firstconnect = false;
    MainWindow::on_actionScan_triggered();
}

MainWindow::~MainWindow()
{
    do_usb_close();
    delete ui;
}

void MainWindow::on_actionScan_triggered()
{
    do_usb_close();
    ui->statusBar->clearMessage();
    do_usb_scan();
    if (chip_found) {
        config_bob[0x13] = CAPLOAD_DEFAULT;
        if (i2c_write_byte(CY22150_ADDR, 0x13, config_bob[0x13]) < 0) do_usb_error_display();
        fillCheckboxes();
        fillComboboxes();
        cy_pll.d1 = config_bob[0x0C] & 0x7F;
        cy_pll.d2 = config_bob[0x47] & 0x7F;
        ui->leF1->setText("0.0");
        ui->leF2->setText("0.0");
        computePLLfrequency();
    }
}

void MainWindow::on_actionOpen_triggered()
{
    do_usb_close();
    ui->statusBar->clearMessage();
    if (do_usb_open() >= 0) {
        fillCheckboxes();
        fillComboboxes();
        cy_pll.d1 = config_bob[0x0C] & 0x7F;
        cy_pll.d2 = config_bob[0x47] & 0x7F;
        computePLLfrequency();
    }
}

void MainWindow::fileread()
{
    QFile iofile(filename);
    unsigned int cnt;
    int flags = 0;
    double f1_in, f2_in;
    QString fstr;
    QByteArray ba_config_bob(256, 0);

    if (!iofile.open(QIODevice::ReadOnly)) return;
    QDataStream in(&iofile);
    in >> ba_config_bob;
    in >> Fref;
    in >> flags;
    in >> f1_in;
    in >> f2_in;
    in >> cy_pll.f1;
    in >> cy_pll.f2;
    iofile.close();

    do_not_change_sem = true;

    for (cnt=0; cnt<sizeof(config_bob); cnt++) config_bob[cnt] = ba_config_bob.data()[cnt];
    ui->sbFreference->setValue(int(Fref));
    if (flags & (1 << 3)) {
        if (!ui->cbPLL_extend->isChecked()) ui->cbPLL_extend->toggle();
    } else {
        if (ui->cbPLL_extend->isChecked()) ui->cbPLL_extend->toggle();
    }
    ui->cbF1_pref->setChecked(flags & (1 << 2));
    ui->cbF2_pref->setChecked(flags & (1 << 1));
    ui->cbF2_enable->setChecked(flags & 1);
    ui->leF2->setEnabled(flags & 1);
    ui->cbF2_pref->setEnabled(flags & 1);
    fstr.sprintf("%12.5f", f1_in);
    ui->leF1->setText(fstr);
    fstr.sprintf("%12.5f", f2_in);
    ui->leF2->setText(fstr);
    ui->statusBar->showMessage("Info: " + filename + " successfully loaded.", 0);
    fillCheckboxes();
    fillComboboxes();
    computePLLfrequency();
    writeALLconfig();

    do_not_change_sem = false;
}

void MainWindow::filewrite()
{
    QFile iofile(filename);
    int flags = 0;
    bool dummy;
    QByteArray ba_config_bob = QByteArray::fromRawData(config_bob, sizeof(config_bob));

    if (!iofile.open(QIODevice::WriteOnly)) return;
    QDataStream out(&iofile);

    out << ba_config_bob;
    out << Fref;
    if (ui->cbPLL_extend->isChecked()) flags |= 1 << 3;
    if (ui->cbF1_pref->isChecked()) flags |= 1 << 2;
    if (ui->cbF2_pref->isChecked()) flags |= 1 << 1;
    if (ui->cbF2_enable->isChecked()) flags |= 1;
    out << flags;
    out << ui->leF1->text().toDouble(&dummy);          // requested output divider 1
    out << ui->leF2->text().toDouble(&dummy);          // requested output divider 2
    out << cy_pll.f1;
    out << cy_pll.f2;
    iofile.close();
    ui->statusBar->showMessage("Info: " + filename + " successfully saved.", 0);
}

void MainWindow::on_actionLoad_triggered()
{
    ui->statusBar->clearMessage();
    filename = QFileDialog::getOpenFileName(this, tr("Load File..."), QDir::homePath(), tr("CY22150 files (*.cy22150.bin)"));
    if (filename.isEmpty()) {
        ui->statusBar->showMessage("Error: Filename is empty!", 0);
        return;
    }
    fileread();
}

void MainWindow::on_actionSave_triggered()
{
    ui->statusBar->clearMessage();
    if (filename.isEmpty())
        filename = QFileDialog::getSaveFileName(this, tr("Save File..."), QDir::homePath(), tr("CY22150 files (*.cy22150.bin)"));
    if (filename.isEmpty()) {
        ui->statusBar->showMessage("Error: Filename is empty!", 0);
        return;
    }
    if (!(filename.endsWith(".cy22150.bin", Qt::CaseInsensitive))) filename += ".cy22150.bin";
    filewrite();
}

void MainWindow::on_actionSave_As_triggered()
{
    ui->statusBar->clearMessage();
    filename = QFileDialog::getSaveFileName(this, tr("Save File As..."), QDir::homePath(), tr("CY22150 files (*.cy22150.bin)"));
    if (filename.isEmpty()) {
        ui->statusBar->showMessage("Error: Filename is empty!", 0);
        return;
    }
    if (!(filename.endsWith(".cy22150.bin", Qt::CaseInsensitive))) filename += ".cy22150.bin";
    filewrite();
}

void MainWindow::fillCheckboxes()
{
    int cnt, bit_cnt, cidx;
    unsigned char cbshift;
    char tmps[256];
    bool conv_ok;
    QLabel *addrLabel;
    QLineEdit *lineEdit;
    QCheckBox *checkBox = NULL;

    computePLLfrequency();

    for (cnt = 1; cnt < ui->gridLayout->rowCount(); cnt += 3) {
        addrLabel = (QLabel *)ui->gridLayout->itemAtPosition(cnt, 1)->widget();
        lineEdit = (QLineEdit *)ui->gridLayout->itemAtPosition(cnt, 2)->widget();
        cidx = addrLabel->text().toInt(&conv_ok, 16);
        snprintf(tmps, sizeof(tmps), "0x%02X", (unsigned char)config_bob[cidx]);
        lineEdit->setText(tmps);
        cbshift = config_bob[cidx];
        for (bit_cnt = ui->gridLayout->columnCount() - 1; bit_cnt > ui->gridLayout->columnCount() - 9; bit_cnt--) {
            checkBox = (QCheckBox *)ui->gridLayout->itemAtPosition(cnt, bit_cnt)->widget();
            if (!firstconnect) checkBox->disconnect(checkBox, SIGNAL(toggled(bool)), this, SLOT(on_anycheckbox_toggled()));
            if (cbshift & 1) {
                checkBox->setChecked(true);
            } else {
                checkBox->setChecked(false);
            }
            cbshift >>= 1;
            checkBox->connect(checkBox, SIGNAL(toggled(bool)), this, SLOT(on_anycheckbox_toggled()));
        }
        if (i2c_write_byte(CY22150_ADDR, (char)cidx, config_bob[cidx]) < 0) do_usb_error_display();
    }
}

void MainWindow::fillComboboxes()
{
    int cnt;
    unsigned long matrixbits = 0;
    QPushButton *clkButton;
    QComboBox *comboBox;

    matrixbits |= ((unsigned char)config_bob[0x46] >> 6) | ((unsigned char)config_bob[0x45] << 2) | ((unsigned char)config_bob[0x44] << 10);
    for (cnt = 0; cnt < ((ui->gridLayout_2->rowCount() + 1) / 3); cnt++) {
        comboBox = (QComboBox *)ui->gridLayout_2->itemAtPosition(3 * cnt, 0)->widget();
        if (!firstconnect) comboBox->disconnect(comboBox, SIGNAL(currentIndexChanged(int)), this, SLOT(on_anymatrixcombobox_changed()));
        comboBox->setCurrentIndex((matrixbits >> (3 * (5 - cnt)) & 7));
        comboBox->connect(comboBox, SIGNAL(currentIndexChanged(int)), this, SLOT(on_anymatrixcombobox_changed()));

        clkButton = (QPushButton *)ui->gridLayout_2->itemAtPosition(3 * cnt, 2)->widget();
        clkButton->setChecked(config_bob[0x09] & (1 << cnt));
        if (firstconnect) clkButton->connect(clkButton, SIGNAL(toggled(bool)), this, SLOT(on_anyclockbutton_toggled()));
    }

    if (!firstconnect) ui->sbXCOdrive->disconnect(ui->sbXCOdrive, SIGNAL(valueChanged(int)), this, SLOT(on_anysbvalue_changed()));
    ui->sbXCOdrive->setValue(((unsigned char)config_bob[0x12] >> 3) & 3);
    ui->sbXCOdrive->connect(ui->sbXCOdrive, SIGNAL(valueChanged(int)), this, SLOT(on_anysbvalue_changed()));

    if (!firstconnect) ui->sbCAPLOAD->disconnect(ui->sbCAPLOAD, SIGNAL(valueChanged(int)), this, SLOT(on_anysbvalue_changed()));
    ui->sbCAPLOAD->setValue((unsigned char)config_bob[0x13]);
    ui->sbCAPLOAD->connect(ui->sbCAPLOAD, SIGNAL(valueChanged(int)), this, SLOT(on_anysbvalue_changed()));

    if (!firstconnect) ui->sbChargePump->disconnect(ui->sbChargePump, SIGNAL(valueChanged(int)), this, SLOT(on_anysbvalue_changed()));
    ui->sbChargePump->setValue(((unsigned char)config_bob[0x40] >> 2) & 7);
    ui->sbChargePump->connect(ui->sbChargePump, SIGNAL(valueChanged(int)), this, SLOT(on_anysbvalue_changed()));

    if (!firstconnect) ui->sbPB->disconnect(ui->sbPB, SIGNAL(valueChanged(int)), this, SLOT(on_anysbvalue_changed()));
    ui->sbPB->setValue((((unsigned char)config_bob[0x40] & 3) << 8) | (unsigned char)config_bob[0x41]);
    ui->sbPB->connect(ui->sbPB, SIGNAL(valueChanged(int)), this, SLOT(on_anysbvalue_changed()));

    if (!firstconnect) ui->sbQCounter->disconnect(ui->sbQCounter, SIGNAL(valueChanged(int)), this, SLOT(on_anysbvalue_changed()));
    ui->sbQCounter->setValue((unsigned char)config_bob[0x42] & 0x7F);
    ui->sbQCounter->connect(ui->sbQCounter, SIGNAL(valueChanged(int)), this, SLOT(on_anysbvalue_changed()));

    if (!firstconnect) ui->sbDIV1N->disconnect(ui->sbDIV1N, SIGNAL(valueChanged(int)), this, SLOT(on_anysbvalue_changed()));
    div1n = (unsigned char)config_bob[0x0C] & 0x7F;
    ui->sbDIV1N->setValue(div1n);
    ui->sbDIV1N->connect(ui->sbDIV1N, SIGNAL(valueChanged(int)), this, SLOT(on_anysbvalue_changed()));

    if (!firstconnect) ui->sbDIV2N->disconnect(ui->sbDIV2N, SIGNAL(valueChanged(int)), this, SLOT(on_anysbvalue_changed()));
    div2n = (unsigned char)config_bob[0x47] & 0x7F;
    ui->sbDIV2N->setValue(div2n);
    ui->sbDIV2N->connect(ui->sbDIV2N, SIGNAL(valueChanged(int)), this, SLOT(on_anysbvalue_changed()));

    ui->cbPO->setChecked(config_bob[0x42] & 0x80);
    ui->rbDIV1CLK_ref->setChecked(config_bob[0x0C] & 0x80);
    ui->rbDIV1CLK_PLL->setChecked(!(config_bob[0x0C] & 0x80));
    ui->rbDIV2CLK_ref->setChecked(config_bob[0x47] & 0x80);
    ui->rbDIV2CLK_PLL->setChecked(!(config_bob[0x47] & 0x80));

    computePLLfrequency();
}

void MainWindow::on_anycheckbox_toggled()
{
    int cnt, bit_cnt, cidx;
    unsigned char cbshift;
    char tmps[256];
    bool conv_ok;
    QLabel *addrLabel;
    QLineEdit *lineEdit;
    QCheckBox *checkBox;

    if (do_not_change_sem) return;

    for (cnt = 1; cnt < ui->gridLayout->rowCount(); cnt += 3) {
        addrLabel = (QLabel *)ui->gridLayout->itemAtPosition(cnt, 1)->widget();
        lineEdit = (QLineEdit *)ui->gridLayout->itemAtPosition(cnt, 2)->widget();
        cidx = addrLabel->text().toInt(&conv_ok, 16);

        cbshift = 0;
        for (bit_cnt = ui->gridLayout->columnCount() - 1; bit_cnt > ui->gridLayout->columnCount() - 9; bit_cnt--) {
            checkBox = (QCheckBox *)ui->gridLayout->itemAtPosition(cnt, bit_cnt)->widget();
            cbshift >>= 1;
            if (checkBox->isChecked() == true) cbshift |= 0x80;
        }
        config_bob[cidx] = cbshift;
        snprintf(tmps, sizeof(tmps), "0x%02X", (unsigned char)config_bob[cidx]);
        if (lineEdit->text() != tmps) {
            lineEdit->setText(tmps);
            if (i2c_write_byte(CY22150_ADDR, (char)cidx, config_bob[cidx]) < 0) do_usb_error_display();
        }
    }
    fillComboboxes();
}

void MainWindow::on_anyclockbutton_toggled()
{
    int cnt;
    unsigned char cbshift;
    QPushButton *clkButton;

    if (do_not_change_sem) return;

    cbshift = 0;
    for (cnt = 0; cnt < ((ui->gridLayout_2->rowCount() + 1) / 3); cnt++) {
        clkButton = (QPushButton *)ui->gridLayout_2->itemAtPosition(3 * cnt, 2)->widget();
        cbshift >>= 1;
        if (clkButton->isChecked() == true) cbshift |= 0x20;
    }
    config_bob[0x09] &= 0xC0;
    config_bob[0x09] |= cbshift;
    fillCheckboxes();
    if (i2c_write_byte(CY22150_ADDR, 0x09, config_bob[0x09]) < 0) do_usb_error_display();
}

void MainWindow::on_anymatrixcombobox_changed()
{
    int cnt;
    unsigned long matrixbits = 0;
    QComboBox *comboBox;

    if (do_not_change_sem) return;

    for (cnt = 0; cnt < ((ui->gridLayout_2->rowCount() + 1) / 3); cnt++) {
        comboBox = (QComboBox *)ui->gridLayout_2->itemAtPosition(3 * cnt, 0)->widget();
        matrixbits <<= 3;
        matrixbits |= (unsigned char)(comboBox->currentIndex() & 7);
    }
    config_bob[0x44] = (matrixbits >> 10) & 0xFF;
    config_bob[0x45] = (matrixbits >> 2) & 0xFF;
    config_bob[0x46] = ((matrixbits << 6) & 0xFF) | (config_bob[0x46] & 0x3F);
    for (cnt = 0x44; cnt < 0x47; cnt++) {
        if (i2c_write_byte(CY22150_ADDR, cnt, config_bob[cnt]) < 0) {
            do_usb_error_display();
            return;
        }
    }
    fillCheckboxes();
}

void MainWindow::on_anysbvalue_changed()
{
    if (do_not_change_sem) return;

    div1n = ui->sbDIV1N->value();
    config_bob[0x0C] &= 0x80;
    config_bob[0x0C] |= div1n & 0x7F;
    div2n = ui->sbDIV2N->value();
    config_bob[0x47] &= 0x80;
    config_bob[0x47] |= div2n & 0x7F;
    config_bob[0x12] &= 0xE7;
    config_bob[0x12] |= (ui->sbXCOdrive->value() & 3) << 3;
    config_bob[0x13] = ui->sbCAPLOAD->value();
    config_bob[0x40] &= 0xE3;
    config_bob[0x40] |= (ui->sbChargePump->value() & 7) << 2;
    config_bob[0x40] &= 0xFC;
    config_bob[0x40] |= (ui->sbPB->value() >> 8) & 3;
    config_bob[0x41] = ui->sbPB->value() & 0xFF;
    config_bob[0x42] &= 0x80;
    config_bob[0x42] |= ui->sbQCounter->value() & 0x7F;
    fillCheckboxes();
    writeALLconfig();
}

void MainWindow::on_cbPO_clicked()
{
    if (ui->cbPO->isChecked()) {
        config_bob[0x42] |= 0x80;
    } else {
        config_bob[0x42] &= 0x7F;
    }
    fillCheckboxes();
    if (i2c_write_byte(CY22150_ADDR, 0x42, config_bob[0x42]) < 0) do_usb_error_display();
}

void MainWindow::on_rbDIV1CLK_ref_clicked()
{
    ui->rbDIV1CLK_ref->toggle();
    if (ui->rbDIV1CLK_ref->isChecked()) {
        config_bob[0x0C] |= 0x80;
    } else {
        config_bob[0x0C] &= 0x7F;
    }
    fillCheckboxes();
    if (i2c_write_byte(CY22150_ADDR, 0x0C, config_bob[0x0C]) < 0) do_usb_error_display();
}

void MainWindow::on_rbDIV1CLK_PLL_clicked()
{
    ui->rbDIV1CLK_PLL->toggle();
    if (ui->rbDIV1CLK_ref->isChecked()) {
        config_bob[0x0C] |= 0x80;
    } else {
        config_bob[0x0C] &= 0x7F;
    }
    fillCheckboxes();
    if (i2c_write_byte(CY22150_ADDR, 0x0C, config_bob[0x0C]) < 0) do_usb_error_display();
}

void MainWindow::on_rbDIV2CLK_ref_clicked()
{
    ui->rbDIV2CLK_ref->toggle();
    if (ui->rbDIV2CLK_ref->isChecked()) {
        config_bob[0x47] |= 0x80;
    } else {
        config_bob[0x47] &= 0x7F;
    }
    fillCheckboxes();
    if (i2c_write_byte(CY22150_ADDR, 0x47, config_bob[0x47]) < 0) do_usb_error_display();
}

void MainWindow::on_rbDIV2CLK_PLL_clicked()
{
    ui->rbDIV2CLK_PLL->toggle();
    if (ui->rbDIV2CLK_ref->isChecked()) {
        config_bob[0x47] |= 0x80;
    } else {
        config_bob[0x47] &= 0x7F;
    }
    fillCheckboxes();
    if (i2c_write_byte(CY22150_ADDR, 0x47, config_bob[0x47]) < 0) do_usb_error_display();
}

void MainWindow::on_leF1_editingFinished()
{
    char tmps[256];
    bool conv_f1_ok, conv_f2_ok;
    QPalette leF1Pal = ui->leF1->palette();

    if (do_not_change_sem) return;

    if (!ui->cbF2_enable->isChecked()) {
        ui->leF2->setText(ui->leF1->text());
        cy_pll.d2 = ui->sbDIV2N->value();
    }

    cy_pll.f_ref = Fref;                                        // reference input frequency
    cy_pll.f1_importance = ui->cbF1_pref->isChecked() ? 1 : 0;  // wichtigkeit, 1=wichtig
    cy_pll.f2_importance = ui->cbF2_pref->isChecked() ? 1 : 0;
    cy_pll.f1 = ui->leF1->text().trimmed().toDouble(&conv_f1_ok);          // requested output divider 1
    cy_pll.f2 = ui->leF2->text().trimmed().toDouble(&conv_f2_ok);          // requested output divider 2
    if (!ui->cbF2_enable->isChecked() || !conv_f2_ok) cy_pll.f2 = cy_pll.f1;

    snprintf(tmps, sizeof(tmps), "%12.5f", cy_pll.f1);
    ui->leF1->setText(tmps);
    snprintf(tmps, sizeof(tmps), "%12.5f", cy_pll.f2);
    ui->leF2->setText(tmps);

    if (conv_f1_ok) {
        findPLLsettings(&cy_pll);
        if (cy_pll.frac < 0) {
            leF1Pal.setBrush(ui->leF1->backgroundRole(), Qt::red);
        } else {
            leF1Pal.setBrush(ui->leF1->backgroundRole(), Qt::white);
        }
        ui->leF1->setPalette(leF1Pal);
        fillComboboxes();
        fillCheckboxes();
        computePLLfrequency();
        writeALLconfig();
    }
}

void MainWindow::on_leF2_editingFinished()
{
    char tmps[256];
    bool conv_f1_ok, conv_f2_ok;
    QPalette leF2Pal = ui->leF2->palette();

    if (do_not_change_sem) return;

    cy_pll.f_ref = Fref;                                        // reference input frequency
    cy_pll.f1_importance = ui->cbF1_pref->isChecked() ? 1 : 0;  // wichtigkeit, 1=wichtig
    cy_pll.f2_importance = ui->cbF2_pref->isChecked() ? 1 : 0;
    cy_pll.f1 = ui->leF1->text().trimmed().toDouble(&conv_f1_ok);          // requested output divider 1
    cy_pll.f2 = ui->leF2->text().trimmed().toDouble(&conv_f2_ok);          // requested output divider 2
    if (!conv_f1_ok) cy_pll.f1 = cy_pll.f2;

    snprintf(tmps, sizeof(tmps), "%12.5f", cy_pll.f1);
    ui->leF1->setText(tmps);
    snprintf(tmps, sizeof(tmps), "%12.5f", cy_pll.f2);
    ui->leF2->setText(tmps);

    if (conv_f2_ok) {
        findPLLsettings(&cy_pll);
        if (cy_pll.frac < 0) {
            leF2Pal.setBrush(ui->leF2->backgroundRole(), Qt::red);
        } else {
            leF2Pal.setBrush(ui->leF2->backgroundRole(), Qt::white);
        }
        ui->leF2->setPalette(leF2Pal);
        fillComboboxes();
        fillCheckboxes();
        computePLLfrequency();
        writeALLconfig();
    }
}

void MainWindow::on_cbF2_enable_toggled(bool checked)
{
    if (do_not_change_sem) return;

    do_not_change_sem = true;
    ui->leF2->setEnabled(checked);
    if (!checked) {
        ui->leF2->setText(ui->leF1->text());
        ui->cbF2_pref->setChecked(false);
    }
    ui->cbF2_pref->setEnabled(checked);
    do_not_change_sem = false;
    MainWindow::on_leF2_editingFinished();
}

void MainWindow::on_cbPLL_extend_toggled(bool checked)
{
    if (checked) {
        f_pll_min = 25000000;	// 25 MHz
        f_pll_max = 450000000;	// 450 MHz
    } else {
        f_pll_min = 100000000;	// 100 MHz
        f_pll_max = 400000000;	// 400 MHz
    }
}

void MainWindow::on_leF1_returnPressed()
{
    ui->leF1->setText(ui->leF1->text().trimmed());
}


void MainWindow::on_leF2_returnPressed()
{
    ui->leF2->setText(ui->leF2->text().trimmed());
}

void MainWindow::on_cbF1_pref_clicked()
{
//    ui->cbF1_pref->toggle();
    MainWindow::on_leF1_editingFinished();
}

void MainWindow::on_cbF2_pref_clicked()
{
//    ui->cbF2_pref->toggle();
    MainWindow::on_leF2_editingFinished();
}

void MainWindow::on_sbFreference_valueChanged(int )
{
    Fref = ui->sbFreference->value();
    computePLLfrequency();
}

void MainWindow::computeFrequencies()
{
    int cnt;
    char tmps[256];
    bool dummy;
    double div1freq, div2freq, resultfreq = 0, lef1val, lef2val;
    QLineEdit *lineEdit;
    QComboBox *comboBox;
    QPalette leF1Pal = ui->leF1->palette();
    QPalette leF2Pal = ui->leF2->palette();

    if (ui->rbDIV1CLK_ref->isChecked()) {
        div1freq = Fref;
    } else {
        div1freq = Fpll;
    }
    if (ui->rbDIV2CLK_ref->isChecked()) {
        div2freq = Fref;
    } else {
        div2freq = Fpll;
    }

    for (cnt = 0; cnt < ((ui->gridLayout_2->rowCount() + 1) / 3); cnt++) {
        comboBox = (QComboBox *)ui->gridLayout_2->itemAtPosition(3 * cnt, 0)->widget();
        lineEdit = (QLineEdit *)ui->gridLayout_2->itemAtPosition(3 * cnt + 1, 0)->widget();
        switch(comboBox->currentIndex()) {
            case 0:
                resultfreq = Fref;
                break;
            case 1:
                resultfreq = div1freq / div1n;
                break;
            case 2:
                resultfreq = div1freq / 2;
                break;
            case 3:
                resultfreq = div1freq / 3;
                break;
            case 4:
                resultfreq = div2freq / div2n;
                break;
            case 5:
                resultfreq = div2freq / 2;
                break;
            case 6:
                resultfreq = div2freq / 4;
                break;
            case 7:
                resultfreq = 0;
                break;
        }
        snprintf(tmps, sizeof(tmps), "%12.5f Hz", resultfreq);
        lineEdit->setText(tmps);
        do_not_change_sem = true;
        if (cnt == 0) {
            snprintf(tmps, sizeof(tmps), "%12.5f", resultfreq);
            lef1val = ui->leF1->text().trimmed().toDouble(&dummy);
            if (lef1val == 0.0) {
                ui->leF1->setText(tmps);
                lef1val = resultfreq;
            }
            if (resultfreq / lef1val > 1.2 || resultfreq / lef1val < 0.833333) leF1Pal.setBrush(ui->leF1->backgroundRole(), Qt::red);
            if (resultfreq / lef1val <= 1.2 && resultfreq / lef1val >= 0.833333) leF1Pal.setBrush(ui->leF1->backgroundRole(), Qt::yellow);
            if (fabs(resultfreq - lef1val) < 0.00002) leF1Pal.setBrush(ui->leF1->backgroundRole(), Qt::green);
            ui->leF1->setPalette(leF1Pal);
        }
        if (cnt == 1) {
            snprintf(tmps, sizeof(tmps), "%12.5f", resultfreq);
            lef2val = ui->leF2->text().trimmed().toDouble(&dummy);
            if (lef2val == 0.0) {
                ui->leF2->setText(tmps);
                lef2val = resultfreq;
            }
            if (ui->cbF2_enable->isChecked()) {
                if (resultfreq / lef2val > 1.2 || resultfreq / lef2val < 0.833333) leF2Pal.setBrush(ui->leF2->backgroundRole(), Qt::red);
                if (resultfreq / lef2val <= 1.2 && resultfreq / lef2val >= 0.833333) leF2Pal.setBrush(ui->leF2->backgroundRole(), Qt::yellow);
                if (fabs(resultfreq - lef2val) < 0.00002) leF2Pal.setBrush(ui->leF2->backgroundRole(), Qt::green);
                ui->leF2->setPalette(leF2Pal);
            }
        }
        do_not_change_sem = false;
    }
}

void MainWindow::computePLLfrequency()
{
    char tmps[256];

    ui->statusBar->clearMessage();
    Q = ((unsigned char)config_bob[0x42] & 0x7F) + 2;
    PB = (((unsigned char)config_bob[0x40] & 3) << 8) | (unsigned char)config_bob[0x41];
    PO = ui->cbPO->isChecked() ? 1 : 0;
    P = (2 * (PB + 4)) + PO;
    Fpll = P * ui->sbFreference->value() / Q;
    snprintf(tmps, sizeof(tmps), "%12.5f Hz", Fpll);
    ui->lePLL->setText(tmps);
    if (Fpll < f_pll_min) ui->statusBar->showMessage("Warning: unstable LOW PLL frequency!", 0);
    if (Fpll > f_pll_max) ui->statusBar->showMessage("Warning: unstable HIGH PLL frequency!", 0);
    computeFrequencies();
}

void MainWindow::findPLLsettings(pll_settings *pll_vars)
{
    #define d_round     0.5
    #define qmin        2               // q range
    #define qmaxmax     129
    #define qmaxf       250000
    #define pmin        8               // p range
    #define pmax        2055
    #define dmin        2               // divider range
    #define dmax        128

    // precalculated constant values
    double f1f = pll_vars->f1_importance ? 10000.0 : 1;
    double f2f = pll_vars->f2_importance ? 10000.0 : 1;
    double pminf = f_pll_min / pll_vars->f_ref;
    double qminf = pll_vars->f_ref / f_pll_max;
    double qmax = pll_vars->f_ref / qmaxf;
    if (qmax > qmaxmax) qmax = qmaxmax;
    double f1bak = pll_vars->f1;
    double f2bak = pll_vars->f2;

    // start conditions
    double bestfrac = -1;
    double frac = -1;
    pll_vars->frac = -1;
    double q = qmin;
    double p = int(q * pminf);
    if (p < pmin) p = pmin ;
    int n = 0;

    if (!ui->cbF2_enable->isChecked()) pll_vars->f2 = pll_vars->f1;

    if (pll_vars->f1 < f_pll_min / 127 && pll_vars->f2 > f_pll_min / 127) pll_vars->f1 = pll_vars->f2;
    if (pll_vars->f2 < f_pll_min / 127 && pll_vars->f1 > f_pll_min / 127) pll_vars->f2 = pll_vars->f1;

    while (1) {
        n++;
        if (q > qmax) break;

        double d1 = (pll_vars->f_ref * p) / (pll_vars->f1 * q) + d_round;
        double d2 = (pll_vars->f_ref * p) / (pll_vars->f2 * q) + d_round;

        if (p > pmax || q < p * qminf || d1 >= dmax || d2 >= dmax) {
            q++;
            p = int(q * pminf);
            if (p < pmin) p = pmin;
            continue;
        }

        if (p < q * pminf || d1 < dmin || d2 < dmin || int(d2) == 3) {
            p++;
            continue;
        }

        double f1t = pll_vars->f_ref * p / (q * int(d1));
        double f2t = pll_vars->f_ref * p / (q * int(d2));
        if (ui->cbF2_enable->isChecked()) frac = fabs(pll_vars->f1 - f1t) * f1f + fabs(pll_vars->f2 - f2t) * f2f;
        if (!ui->cbF2_enable->isChecked()) frac = fabs(pll_vars->f1 - f1t) * f1f;
        if (bestfrac < 0 || frac < bestfrac) {
            bestfrac = frac;
            pll_vars->frac = frac;
            pll_vars->d1 = int(d1);
            if (ui->cbF2_enable->isChecked()) pll_vars->d2 = int(d2);
            pll_vars->p = int(p);
            pll_vars->q = int(q);
            if (!frac) break;
        }
        p++;
    }
    pll_vars->iterations = n;
    char tmps[256];
    snprintf(tmps, sizeof(tmps), "Iterations: %d", n);
    ui->lbIterations->setText(tmps);

    pll_vars->f1 = f1bak;
    pll_vars->f2 = f2bak;

    if (pll_vars->f1 < f_pll_min / 127) {
        if (!ui->rbDIV1CLK_ref->isChecked()) ui->rbDIV1CLK_ref->toggle();
        config_bob[0x0C] |= 0x80;
        pll_vars->d1 = int(round(pll_vars->f_ref / pll_vars->f1));
        if (pll_vars->d1 > 127) pll_vars->d1 = 127;
    } else {
        if (!ui->rbDIV1CLK_PLL->isChecked()) ui->rbDIV1CLK_PLL->toggle();
        config_bob[0x0C] &= 0x7F;
    }
    div1n = pll_vars->d1;

    if (ui->cbF2_enable->isChecked()) {
        if (pll_vars->f2 < f_pll_min / 127) {
            if (!ui->rbDIV2CLK_ref->isChecked()) ui->rbDIV2CLK_ref->toggle();
            config_bob[0x47] |= 0x80;
            pll_vars->d2 = int(round(pll_vars->f_ref / pll_vars->f2));
            if (pll_vars->d2 > 127) pll_vars->d2 = 127;
        } else {
            if (!ui->rbDIV2CLK_PLL->isChecked()) ui->rbDIV2CLK_PLL->toggle();
            config_bob[0x47] &= 0x7F;
        }
        div2n = pll_vars->d2;
    }
    setPQD1D2(pll_vars);
    setChargePump(pll_vars);
}

void MainWindow::setPQD1D2(pll_settings *pll_vars)
{
    config_bob[0x0C] &= 0x80;
    if (pll_vars->d1 < 4) {
        if (pll_vars->d1 == 2) {
            config_bob[0x0C] |= 4;
            ui->cbMatrix_LCLK1->setCurrentIndex(2);
        }
        if (pll_vars->d1 == 3) {
            config_bob[0x0C] |= 6;
            ui->cbMatrix_LCLK1->setCurrentIndex(3);
        }
    } else {
        config_bob[0x0C] |= pll_vars->d1 & 0x7F;
        ui->cbMatrix_LCLK1->setCurrentIndex(1);
    }

    if (ui->cbF2_enable->isChecked()) {
        config_bob[0x47] &= 0x80;
        if (pll_vars->d2 < 5) {
            if (pll_vars->d2 == 2) {
                config_bob[0x47] |= 4;
                ui->cbMatrix_LCLK2->setCurrentIndex(5);
            }
            if (pll_vars->d2 == 4) {
                config_bob[0x47] |= 8;
                ui->cbMatrix_LCLK2->setCurrentIndex(6);
            }
        } else {
            config_bob[0x47] |= pll_vars->d2 & 0x7F;
            ui->cbMatrix_LCLK2->setCurrentIndex(4);
        }
    }

    if (pll_vars->frac >= 0) {
        int pb;
        // (2*(PB+4)+PO) * Fref / (Q+2)
        pb = (pll_vars->p >> 1) - 4;
        if (pll_vars->p & 1) {
            config_bob[0x42] |= 0x80;
        } else {
            config_bob[0x42] &= 0x7F;
        }
        config_bob[0x40] &= 0xFC;
        config_bob[0x40] |= (pb >> 8) & 3;
        config_bob[0x41] = pb & 0xFF;
        config_bob[0x42] &= 0x80;
        config_bob[0x42] |= (pll_vars->q - 2) & 0x7F;
    }
}

void MainWindow::setChargePump(pll_settings *pll_vars)
{
    config_bob[0x40] &= 0xE3;
    switch (pll_vars->p) {
        case 0 ... 15:
            ui->statusBar->showMessage("Warning: unstable charge-pump-setting, p < 15!", 0);
            break;
        case 16 ... 44:
            break;
        case 45 ... 479:
            config_bob[0x40] |= 1 << 2;
            break;
        case 480 ... 639:
            config_bob[0x40] |= 2 << 2;
            break;
        case 640 ... 799:
            config_bob[0x40] |= 3 << 2;
            break;
        case 800 ... 1023:
            config_bob[0x40] |= 4 << 2;
            break;
        case 1024 ... 1319:
            ui->statusBar->showMessage("Warning: unstable charge-pump-setting, p > 1023!", 0);
            config_bob[0x40] |= 5 << 2;
            break;
        case 1320 ... 1679:
            ui->statusBar->showMessage("Warning: unstable charge-pump-setting, p > 1320!", 0);
            config_bob[0x40] |= 6 << 2;
            break;
        case 1680 ... 2055:
            ui->statusBar->showMessage("Warning: unstable charge-pump-setting, p > 1680!", 0);
            config_bob[0x40] |= 7 << 2;
            break;
        default:
            ui->statusBar->showMessage("Warning: invalid charge-pump-setting, p out of range!", 0);
            break;
    }

}

void MainWindow::writeALLconfig()
{
    if (i2c_write_byte(CY22150_ADDR, 0x0C, config_bob[0x0C]) < 0) do_usb_error_display();
    if (i2c_write_byte(CY22150_ADDR, 0x12, config_bob[0x12]) < 0) do_usb_error_display();
    if (i2c_write_byte(CY22150_ADDR, 0x13, config_bob[0x13]) < 0) do_usb_error_display();
    if (i2c_write_byte(CY22150_ADDR, 0x40, config_bob[0x40]) < 0) do_usb_error_display();
    if (i2c_write_byte(CY22150_ADDR, 0x41, config_bob[0x41]) < 0) do_usb_error_display();
    if (i2c_write_byte(CY22150_ADDR, 0x42, config_bob[0x42]) < 0) do_usb_error_display();
    if (i2c_write_byte(CY22150_ADDR, 0x47, config_bob[0x47]) < 0) do_usb_error_display();
}
